#include <stdio.h>
#include <linux/i2c-dev.h>
#include <linux/i2c.h>
#include <i2c/smbus.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <math.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdint.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <string.h>
#include <time.h>

#define PWR_MGMT_1   0x6B
#define SMPLRT_DIV   0x19
#define CONFIG       0x1A
#define GYRO_CONFIG  0x1B
#define INT_ENABLE   0x38
#define ACCEL_XOUT_H 0x3B
#define ACCEL_YOUT_H 0x3D
#define ACCEL_ZOUT_H 0x3F
#define GYRO_XOUT_H  0x43
#define GYRO_YOUT_H  0x45
#define GYRO_ZOUT_H  0x47
#define MPU6050_I2C_ADDR 0x68

#define I2C_SMBUS_READ	1
#define I2C_SMBUS_WRITE	0

#define I2C_SMBUS_QUICK		    0
#define I2C_SMBUS_BYTE		    1
#define I2C_SMBUS_BYTE_DATA	    2 
#define I2C_SMBUS_WORD_DATA	    3
#define I2C_SMBUS_PROC_CALL	    4
#define I2C_SMBUS_BLOCK_DATA	    5
#define I2C_SMBUS_I2C_BLOCK_BROKEN  6

int fd;

void error_handling(char *message){
	fputs(message, stderr);
	fputc('\n', stderr);
	exit(1);
}

void I2C_WRITE(int fd, __u8 reg_address, __u8 val) {
	char buf[2];
	if(fd < 0) {
		printf("Error, i2c bus is not available\n");
		exit(1);
	}

	buf[0] = reg_address;
	buf[1] = val;

	if (write(fd, buf, 2) != 2) {
		printf("Error, unable to write to i2c device\n");
		exit(1);
	}

}

short merge_bytes(short MSB, short LSB) {
	return  ((( MSB & 0xFF) << 8) | LSB);
}


short read_data(int fd, int addr){
	union i2c_smbus_data data;
  	struct i2c_smbus_ioctl_data args;

  	args.read_write = I2C_SMBUS_READ;
  	args.command    = addr;
  	args.size       = I2C_SMBUS_BYTE_DATA ;
  	args.data       = &data ;
	
	if(ioctl(fd, I2C_SMBUS, &args)){
		return -1;
	}else{
		return data.byte & 0xFF;
	}
}


short find_base(double *baseGx, double *baseGy){
	double sumGx = 0;
	double sumGy = 0;
	time_t start;
	time_t end;
	double Gyro_x;
	double Gyro_y;

	for(int i = 0; i < 100 ; i++){
	Gyro_x = merge_bytes(read_data(fd, GYRO_XOUT_H),read_data(fd, GYRO_XOUT_H +1));
	Gyro_y = merge_bytes(read_data(fd, GYRO_YOUT_H),read_data(fd, GYRO_YOUT_H +1));
	
	sumGx += Gyro_x/131;
	sumGy += Gyro_y/131;
	}
	*baseGx = sumGx / 100;
	*baseGy = sumGy / 100;
	
}

int main(int argc, char **argv)
{
    int sock;
    struct sockaddr_in serv_addr;
	double result;
    char msg[6];
    char on[2] = "1";
    int str_len;
    int light = 0;
    if (argc != 3)
    {
        printf("Usage : %s <IP> <port>\n", argv[0]);
        exit(1);
    }

	sock = socket(PF_INET, SOCK_STREAM, 0);

	    if (sock == -1)
        error_handling("socket() error");
	    memset(&serv_addr, 0, sizeof(serv_addr));

    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr.s_addr = inet_addr(argv[1]);
    serv_addr.sin_port = htons(atoi(argv[2]));
    if (connect(sock, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) == -1)
        error_handling("connect() error");


	time_t start=0, start1 = 0, start2 = 0; 
	time_t end=0, end1 = 0, end2 = 0;
    float RADIAN_TO_DEGREE = 180 / 3.14159; //가속도->각도 사용
	double DEG_PER_SEC = 32767 / 250; //각속도->각도 사용

	double dt = 0; //자이로센서 적분

	float Acc_x,Acc_y,Acc_z;

	float Gyro_x,Gyro_y,Gyro_z;

	float angleAcX = 0,angleAcY = 0;  // 가속도 각도
	float angleGyX = 0, angleGyY = 0;  // 각속도 각도
	float angleAphX = 0, angleAphY = 0;  //상보필터 각도

	float Ax=0, Ay=0, Az=0;
	double Gx=0, Gy=0, Gz=0;

	float alpha = 0.90;
	float angleBlock = 45.0 / 8;    

	double baseGx = 0;
	double baseGy = 0;
	
    int position[4] = {0, 0, 0, 0};
	int preposition[4] = {0, 0, 0, 0};
	int count[4] = {0, 0, 0, 0};
	int sposition[4] = {0, 0, 0, 0};

	int adapter_nr = 1; 

	char *fileName = "/dev/i2c-1";
	fd = open(fileName, O_RDWR);
	if (fd < 0) {

		exit(1);
	}

    if (ioctl(fd, I2C_SLAVE, MPU6050_I2C_ADDR) < 0) {

		exit(1);
	}


    I2C_WRITE(fd, PWR_MGMT_1, 0x01);
	I2C_WRITE(fd, SMPLRT_DIV, 0x07);
	I2C_WRITE(fd, CONFIG, 0x00);
	I2C_WRITE(fd, GYRO_CONFIG, 24);	
	I2C_WRITE(fd, INT_ENABLE, 0x01);
	start2 = time(NULL);
	find_base(&baseGx, &baseGy); // 각속도로 각도 측정 시 base각도 측정 
	while(1)
	{
		start1 = clock();
		
		Acc_x = merge_bytes(read_data(fd, ACCEL_XOUT_H),read_data(fd, ACCEL_XOUT_H +1));
		Acc_y = merge_bytes(read_data(fd, ACCEL_YOUT_H),read_data(fd, ACCEL_YOUT_H +1));
		Acc_z = merge_bytes(read_data(fd, ACCEL_ZOUT_H),read_data(fd, ACCEL_ZOUT_H +1));

		start = end;

		Gyro_x = merge_bytes(read_data(fd, GYRO_XOUT_H),read_data(fd, GYRO_XOUT_H +1));
		Gyro_y = merge_bytes(read_data(fd, GYRO_YOUT_H),read_data(fd, GYRO_YOUT_H +1));
		Gyro_z = merge_bytes(read_data(fd, GYRO_ZOUT_H),read_data(fd, GYRO_ZOUT_H +1));

		end = clock();

		Gx = Gyro_x/DEG_PER_SEC;
		Gy = Gyro_y/DEG_PER_SEC;
		Gz = Gyro_z/DEG_PER_SEC;

		dt = (double)(end - start)/1000;


		Ax = Acc_x/16384.0;
		Ay = Acc_y/16384.0;
		Az = Acc_z/16384.0;

		angleAcY = atan(-Ax /sqrt(pow(Ay, 2) + pow(Az, 2)));
		angleAcX = atan(Ay / sqrt(pow(Ax, 2) + pow(Az, 2)));
		angleAcY *= RADIAN_TO_DEGREE;
		angleAcX *= RADIAN_TO_DEGREE;

		angleAphX = (angleAphX + ((Gx-baseGx) * dt)) * alpha + angleAcX * (1-alpha);
		angleAphY = (angleAphY + ((Gy-baseGy) * dt)) * alpha + angleAcY * (1-alpha);


		for(int i = 0; i < 4; i ++){
			position[i] = 0;
		}

		if(fabs(angleAphX) < angleBlock * 1){  // x가 5.625 미만
			if(angleAphX < 0){
				position[2] += 0;
				position[3] += 0;
			}else{
				position[0] += 0;
				position[1] += 0;
			}
		}else if(fabs(angleAphX) < angleBlock * 2){ //x가 11.25 미만
			if(angleAphX < 0){
				position[2] += 1;
				position[3] += 1;
			}else{
				position[0] += 1;
				position[1] += 1;
			}
		}else if(fabs(angleAphX) < angleBlock * 3){ //x가 16.875 미만
			if(angleAphX < 0){
				position[2] += 2;
				position[3] += 2;
			}else{
				position[0] += 2;
				position[1] += 2;
			}
		}else if(fabs(angleAphX) < angleBlock * 4){ //x가 22.5 미만
			if(angleAphX < 0){
				position[2] += 3;
				position[3] += 3;
			}else{
				position[0] += 3;
				position[1] += 3;
			}
			
		}else if(fabs(angleAphX) < angleBlock * 5){ //x가 28.125 미만
			if(angleAphX < 0){
				position[2] += 4;
				position[3] += 4;
			}else{
				position[0] += 4;
				position[1] += 4;
			}
		}else if(fabs(angleAphX) < angleBlock * 6){ //x가 33.75 미만
			if(angleAphX < 0){
				position[2] += 5;
				position[3] += 5;
			}else{
				position[0] += 5;
				position[1] += 5;
			}
		}else if(fabs(angleAphX) < angleBlock * 7){ //x가 45 미만
			if(angleAphX < 0){
				position[2] += 6;
				position[3] += 6;
			}else{
				position[0] += 6;
				position[1] += 6;
			}
		}else{										//x가 45 이상
			if(angleAphX < 0){
				position[2] += 7;
				position[3] += 7;
			}else{
				position[0] += 7;
				position[1] += 7;
			}
		}

		if(fabs(angleAphY) < angleBlock * 1){	// y가 5.625 미만
			if(angleAphY >= 0){
				position[0] += 0;
				position[2] += 0;
			}else{
				position[1] += 0;
				position[3] += 0;
			}
		}else if(fabs(angleAphY) < angleBlock * 2){     //y가 11.25 미만
			if(angleAphY >= 0){
				position[0] += 1;
				position[2] += 1;
			}else{
				position[1] += 1;
				position[3] += 1;
			}
		}else if(fabs(angleAphY) < angleBlock * 3){		//y가 16.875 이하
			if(angleAphY >= 0){
				position[0] += 2;
				position[2] += 2;
			}else{
				position[1] += 2;
				position[3] += 2;
			}
		}else if(fabs(angleAphY) < angleBlock * 4){		//y가 22.5 이상
			if(angleAphY >= 0){
				position[0] += 3;
				position[2] += 3;
			}else{
				position[1] += 3;
				position[3] += 3;
			}
		}else if(fabs(angleAphY) < angleBlock * 5){		//y가 22.5 이상
			if(angleAphY >= 0){
				position[0] += 4;
				position[2] += 4;
			}else{
				position[1] += 4;
				position[3] += 4;
			}
		}else if(fabs(angleAphY) < angleBlock * 6){		//y가 22.5 이상
			if(angleAphY >= 0){
				position[0] += 5;
				position[2] += 5;
			}else{
				position[1] += 5;
				position[3] += 5;
			}
		}else if(fabs(angleAphY)< angleBlock * 7){		//y가 22.5 이상
			if(angleAphY >= 0){
				position[0] += 6;
				position[2] += 6;
			}else{
				position[1] += 6;
				position[3] += 6;
			}
		}else{
			if(angleAphY >= 0){
				position[0] += 8;
				position[2] += 8;
			}else{
				position[1] += 8;
				position[3] += 8;
			}
		}

		for(int i = 0; i < 4; i++){
			if(position[i]>=8){  // 더했을 때 8이상이면 그냥 7로 설정
				position[i] = 7;
			}
			if(position[i] != preposition[i])  // 노이즈 보정을 위해서 이전값과 비교
				count[i]++;
			else
				count[i] = 0;
			if(count[i] == 3){ // 연속으로 이전값과 다르면 유효한 값으로 판정하고 값 변경
				preposition[i] = position[i];
				sposition[i] = position[i];
				count[i] = 0;
			}
		}
		
		// printf("\n angleAphX = %.3f      angleAphY = %.3f", angleAphX, angleAphY);
		// printf("\n frontleft = %d   frontright = %d    backleft = %d    backright = %d\n", position[0], position[1], position[2], position[3]);

		delay(100);
		// end1 = clock();
		// end2 = time(NULL);
		// printf("How long time is it? %.3f,  %d\n", (double)(end1-start1), end2-start2);
		snprintf(msg, sizeof(msg), "s%d%d%d%d", sposition[0], sposition[1], sposition[2], sposition[3]);
        write(sock, msg, sizeof(msg));
        printf("msg = %s\n",msg);
		usleep(1000);
    }
	close(sock);
}
